/*
 * Copyright (c) 2003-2005 Erez Zadok
 * Copyright (c) 2003-2005 Charles P. Wright
 * Copyright (c) 2003-2005 Mohammad Nayyer Zubair
 * Copyright (c) 2003-2005 Puja Gupta
 * Copyright (c) 2003-2005 Harikesavan Krishnan
 * Copyright (c) 2003-2005 David P. Quigley
 * Copyright (c) 2003-2005 Stony Brook University
 * Copyright (c) 2003-2005 The Research Foundation of State University of New York
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package.
 *
 * This Copyright notice must be kept intact and distributed with all sources.
 */

#include "fist.h"
#include "unionfs.h"
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <linux/kdev_t.h>
#endif

/*
 * verify_forwardmap(super_block *sb)
 * sb: pointer to a superblock containing the forwardmap.
 * returns: 0 on success EINVAL or ENOMEM on failure;
 */
static int verify_forwardmap(struct super_block *sb)
{
	int err = 0,bytesread = 0,bindex = 0, mallocsize = 0,i = 0;
	loff_t readpos = 0;
	struct file *forwardmap = NULL;
	struct fmaphdr header;
	mm_segment_t oldfs;
	print_entry_location();

	forwardmap = stopd(sb)->usi_forwardmap;
	if(!forwardmap){
		err = -EINVAL;
		goto out;
	}
	oldfs = get_fs();
	set_fs(KERNEL_DS);
	bytesread = forwardmap->f_op->read(forwardmap, (char *)&header, sizeof(struct fmaphdr), &readpos);
	set_fs(oldfs);
	if(bytesread < sizeof(struct fmaphdr)) { 
		err = -EINVAL;
		goto out;
	}
	if(header.magic != FORWARDMAP_MAGIC || header.version != FORWARDMAP_VERSION) {
		err = -EINVAL;
		goto out;
	}
	stopd(sb)->usi_bmap = KMALLOC(sizeof(uuid_t) * header.usedbranches,GFP_KERNEL);
	stopd(sb)->usi_num_bmapents = header.usedbranches;
	if (!stopd(sb)->usi_bmap) {
		err = -ENOMEM;
		goto out;
	}
	oldfs = get_fs();
	set_fs(KERNEL_DS);
	while(bindex < header.usedbranches){
		bytesread = forwardmap->f_op->read(forwardmap,(char *)&stopd(sb)->usi_bmap[bindex],sizeof(uuid_t),&readpos);
		if(bytesread < sizeof(struct fmapent)){ 
			err = -EINVAL;
			set_fs(oldfs);
			goto out_err;
		}
		bindex++;
	}
	set_fs(oldfs);
	mallocsize = sizeof(int) * header.usedbranches;
	stopd(sb)->usi_fsnum_table = KMALLOC(mallocsize,GFP_KERNEL);
	for(i = 0; i < header.usedbranches;i++){
		stopd(sb)->usi_fsnum_table[i] = -1;
	}
	if (!stopd(sb)->usi_fsnum_table) {
		err = -ENOMEM;
		goto out_err;
	}	
	
out_err:
	if(stopd(sb)->usi_bmap){
		KFREE(stopd(sb)->usi_bmap);
	}	
out:
	print_exit_status(err);
	return err;
}

/* 
 * verify_reversemap(struct super_block sb, int rmapindex)
 *
 * sb: The unionfs superblock containing all of the current imap info
 * rmapindex: the index in the usi_reversemaps array that we wish to 
 * verify
 *
 * Assumes the reverse maps less than rmapindex are valid.
 * 
 * returns: 0 if the opperation succeds
 * 	-EINVAL if the map file does not belong to the forward map
 *
 */
static int verify_reversemap(struct super_block *sb,int rmapindex, struct unionfs_dentry_info *hidden_root_info )
{
	int err = 0, i = 0, j =0,found = 0, bytesread;
	loff_t readpos = 0;
	struct file *forwardmap,*reversemap;
	struct fmaphdr fheader;
	struct rmaphdr rheader;
	mm_segment_t oldfs;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	struct statfs st;
#else
	struct kstatfs st;
#endif

	print_entry_location();
	forwardmap = stopd(sb)->usi_forwardmap;
	if(!forwardmap){
		err = -EINVAL;
		goto out;
	}
	reversemap = stopd(sb)->usi_reversemaps[rmapindex];
	if(!reversemap){
		err = -EINVAL;
		goto out;
	}
	oldfs = get_fs();
	set_fs(KERNEL_DS);
	bytesread = forwardmap->f_op->read(forwardmap,(char *)&fheader,sizeof(struct fmaphdr),&readpos);
	if(bytesread < sizeof(struct fmaphdr)){ 
		set_fs(oldfs);	
		err = -EINVAL;
		goto out;
	}
	readpos = 0;
	bytesread = reversemap->f_op->read(reversemap,(char *)&rheader,sizeof(struct rmaphdr),&readpos);
	if(bytesread < sizeof(struct rmaphdr)){ 
		set_fs(oldfs);
		err = -EINVAL;
		goto out;
	}
	set_fs(oldfs);
	if(rheader.magic != REVERSEMAP_MAGIC || rheader.version != REVERSEMAP_VERSION){
		err = -EINVAL;
		goto out;
	}
	if(memcmp(fheader.uuid,rheader.fwduuid, sizeof(fheader.uuid))){
		err = -EINVAL;
		goto out;
	}

	/* XXX: Ok so here we take the new map and read the fsid from it. Then
	 * we go through all the branches in the union and see which ones it
	 * matches with*/
	for(i = 0; i < stopd(sb)->usi_num_bmapents && !found; i++){
		if(memcmp(rheader.revuuid,stopd(sb)->usi_bmap[i],sizeof(rheader.revuuid))) {
			found = 1;
			for(j = 0; j < hidden_root_info->udi_bend; j++) {
				fsid_t fsid;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)	
				kdev_t dev;
				memset(&st,0,sizeof(struct statfs));
#else
				dev_t dev;
				memset(&st,0,sizeof(struct kstatfs));
#endif
				err = stohs_index(sb,j)->s_op->statfs( stohs_index(sb,j),&st);
				if (err){
					goto out;
				}
				if(st.f_fsid.val[0] || st.f_fsid.val[1]){
					fsid = st.f_fsid;
				} else {
					dev = stohs_index(sb,j)->s_dev;
					st.f_fsid.val[0] = MAJOR(dev);
					st.f_fsid.val[1] = MINOR(dev);
				}
				if(!memcmp(&fsid,&rheader.fsid,sizeof(fsid))) {
					if(stopd(sb)->usi_fsnum_table[i] == -1){
						stopd(sb)->usi_fsnum_table[i] = j;
					}
					if(stopd(sb)->usi_map_table[j]){
						printk(KERN_WARNING"Two reverse maps share fsid %u%u!\n", rheader.fsid.val[0], rheader.fsid.val[1]);
						err = -EINVAL;
						goto out;
					} else {
						stopd(sb)->usi_map_table[j] = reversemap;

					}
				}
			}	
		}
	}
	if(!found){
		printk(KERN_WARNING"Could not match the reversemap uuid with an entry in the forwardmap table\n");
		err = -EINVAL;
	}
out:
	print_exit_status(err);
	return err;
}

int parse_imap_options(struct super_block *sb, struct unionfs_dentry_info *hidden_root_info , char** options)
{
	int mallocsize,count = 0,err = 0;
	char *end,*tmp,*name;

	print_entry_location();

	//Make sure our file pointers are set to null initially
	stopd(sb)->usi_forwardmap = NULL;
	stopd(sb)->usi_reversemaps = NULL;
	if (**options == ':') {
		options++;
	}

	/* We don't want to go off the end of our imap arguments later on. */
	for (end = *options; *end; end++) ;

	/* colon at the end */
	if (*(end-1) == ':') {
		end--;
		*end = '\0';
	}

	if (*options == end) {
		err = -EINVAL;
		printk(KERN_WARNING "unionfs: non-blank imap option required\n");
		goto out_error;
	}

	mallocsize = sizeof(struct file *) * ( hidden_root_info->udi_bend + 1);
	stopd(sb)->usi_reversemaps = KMALLOC(mallocsize, GFP_UNIONFS);
	if (IS_ERR(stopd(sb)->usi_reversemaps)) {
		err = -ENOMEM;
		goto out_error;
	}
	memset(stopd(sb)->usi_reversemaps, 0, mallocsize);
	stopd(sb)->usi_map_table = KMALLOC(mallocsize,GFP_UNIONFS);
	if (IS_ERR(stopd(sb)->usi_map_table)) {
		err = -ENOMEM;
		goto out_error;
	}
	memset(stopd(sb)->usi_map_table, 0, mallocsize);
	while (*options < end) {

		tmp = *options;

		while (*tmp && *tmp != ':') {
			tmp++;
		}
		*tmp = '\0';

		/* name contains individual dir */
		name = *options;
		if (!stopd(sb)->usi_forwardmap) {
			stopd(sb)->usi_forwardmap = filp_open(name, O_RDWR, 0);
			if (IS_ERR(stopd(sb)->usi_forwardmap)) {
				err = PTR_ERR(stopd(sb)->usi_forwardmap);
				stopd(sb)->usi_forwardmap = NULL;
				goto out_error;
			}
			err = verify_forwardmap(sb);
			if(err)
				goto out_error;
		}
		else {
			stopd(sb)->usi_reversemaps[count] = filp_open(name, O_RDWR, 0);
			if (IS_ERR(stopd(sb)->usi_reversemaps[count])) {
				err = PTR_ERR(stopd(sb)->usi_reversemaps[count]);
				stopd(sb)->usi_reversemaps[count] = NULL;
				goto out_error;
				
			}
			count++;
			err = verify_reversemap(sb,count-1,hidden_root_info);
			if(err)
				goto out_error;

		}
		/* increment the string pointer */
		*options = tmp + 1;
	}	
	if(err)
		goto out_error;
	
	/* Initialize the super block's next_avail field */
	/* Dave, you can't use 64-bit division here because the i386 doesn't
	 * support it natively.  Instead you need to punt if the size is 
	 * greater than unsigned long, and then cast it down.  Then you should
	 * be able to assign to this value, without having these problems. */

	if(stopd(sb)->usi_forwardmap->f_dentry->d_inode->i_size  > ULONG_MAX){
		return -EFBIG;
	}	
	stopd(sb)->usi_next_avail = ((unsigned long)(stopd(sb)->usi_forwardmap->f_dentry->d_inode->i_size - 
				(sizeof(struct fmaphdr) + sizeof(uuid_t[256])))
			/sizeof(struct fmapent));
	stopd(sb)->usi_num_bmapents = count;
	stopd(sb)->usi_persistent = 1;
	goto out;	

out_error:
	if(stopd(sb)->usi_forwardmap){
		filp_close(stopd(sb)->usi_forwardmap,NULL);
	}
	while(count-1 >= 0) {
		if(stopd(sb)->usi_reversemaps[count]){
			filp_close(stopd(sb)->usi_reversemaps[count],NULL);	
		}
		count--;
	}
out:
	print_exit_status(err);
	return err;
}

/*
 * get_uin(uint8 branchnum, ino_t inode_number, int flag)
 * fsnum: branch to reference when getting the inode number
 * inode_number: lower level inode number use to reference the proper inode.
 * flag: if set to O_CREAT it will creat the entry if it doesent exist
 * 		 otherwise it will return the existing one.
 * returns: the unionfs inode number either created or retrieved based on
 * 			the information.
 */
ino_t get_uin(struct super_block *sb, uint8_t branchnum,ino_t inode_number,int flag) {
	//struct file *reversemap;
	struct file *forwardmap;
	struct fmapent tmp_ent;
	mm_segment_t oldfs;
	int byteswritten = 0;
	ino_t err = 0;
	ino_t uino = -1;

	print_entry_location();
	/* Make sure we have a valid super block */
	PASSERT(sb);
	PASSERT(stopd(sb)->usi_forwardmap);
	PASSERT(stopd(sb)->usi_reversemaps);
	/* Find appropriate reverse map and then read from the required position */
	/* XXX: get it from the array. */

	/* If we haven't found an entry and we have the O_CREAT flag set we want to 
	 * create a new entry write it out to the file and return its index
	 */
	if(uino < 0 && (flag & O_CREAT)) {
		forwardmap = stopd(sb)->usi_forwardmap;
		/* XXX: Fill fsnum from branch->fsnum table */
		tmp_ent.inode = inode_number;
		oldfs = get_fs();
		set_fs(KERNEL_DS);	
		byteswritten = forwardmap->f_op->write(forwardmap,(char *)&tmp_ent,sizeof(struct fmapent),&forwardmap->f_pos);
		set_fs(oldfs);
		if(byteswritten != sizeof(struct fmapent)){
			err = -EIO;
			goto out;
		}
		uino = stopd(sb)->usi_next_avail++;
	}
	else 
		err = -ENOENT;
out:
	if(!err)
		err = uino;

	print_exit_location();
	return err;
}

/*
 * get_lin(ino_t inode_number)
 * inode_number : inode number for the unionfs inode
 * returns: the lower level inode# and branch#
 */
/* entry should use a poiner on the stack. should be staticly allocated one
 * level up*/
int get_lin(struct super_block *sb, ino_t inode_number,struct fmapent *entry) {
	struct file *forwardmap;
	loff_t seek_size;
	mm_segment_t oldfs;
	int err = 0, bytesread = 0;

	print_entry_location();

	PASSERT(sb);
	if(!entry){
		entry = ERR_PTR(-ENOMEM);
		goto out;
	}
	forwardmap = stopd(sb)->usi_forwardmap;
	if(!forwardmap || !forwardmap->f_op->read){
		entry = ERR_PTR(-EINVAL);
		goto out;
	}
	seek_size = sizeof(struct fmaphdr) + sizeof(uuid_t[256]) + (sizeof(struct fmapent) * inode_number);
	oldfs = get_fs();
	set_fs(KERNEL_DS);
	bytesread = forwardmap->f_op->read(forwardmap,(char*)entry,sizeof(struct fmapent),&seek_size);
	set_fs(oldfs);
	if(bytesread != sizeof(struct fmapent)){
		entry = ERR_PTR(-EINVAL);
		goto out;
	}

out:
	print_exit_location();
	return err;
}

/*
 * vim:shiftwidth=8 
 * vim:textwidth=80
 */
